"""SQLAlchemy class representing DBDataset row. Holds dynamic dataset metadata for
each run.
"""
from hdx.database import Base
from sqlalchemy import Boolean, Column, ForeignKey, Integer, String

from . import CustomDateTime
from .dbinfodataset import DBInfoDataset
from .dbrun import DBRun


class DBDataset(Base):
    """
    run_number = Column(
        Integer, ForeignKey(DBRun.run_number), primary_key=True
    )

    id = Column(String, ForeignKey(DBInfoDataset.id), primary_key=True)

    dataset_date = Column(String)

    update_frequency = Column(Integer)

    review_date = Column(CustomDateTime)

    last_modified = Column(CustomDateTime, nullable=False)

    updated_by_script = Column(CustomDateTime)

    metadata_modified = Column(
        CustomDateTime, nullable=False
    )  # this field and above are CKAN fields

    latest_of_modifieds = Column(CustomDateTime, nullable=False)

    what_updated = Column(String, nullable=False)

    last_resource_updated = Column(
        String, nullable=False
    )  # id of last resource updated

    last_resource_modified = Column(
        CustomDateTime, nullable=False
    )  # date last resource updated

    fresh = Column(Integer)

    error = Column(Boolean, nullable=False)
    """

    run_number = Column(
        Integer, ForeignKey(DBRun.run_number), primary_key=True
    )
    id = Column(String, ForeignKey(DBInfoDataset.id), primary_key=True)
    dataset_date = Column(String)
    update_frequency = Column(Integer)
    review_date = Column(CustomDateTime)
    last_modified = Column(CustomDateTime, nullable=False)
    updated_by_script = Column(CustomDateTime)
    metadata_modified = Column(
        CustomDateTime, nullable=False
    )  # this field and above are CKAN fields
    latest_of_modifieds = Column(CustomDateTime, nullable=False)
    what_updated = Column(String, nullable=False)
    last_resource_updated = Column(
        String, nullable=False
    )  # id of last resource updated
    last_resource_modified = Column(
        CustomDateTime, nullable=False
    )  # date last resource updated
    fresh = Column(Integer)
    error = Column(Boolean, nullable=False)

    def __repr__(self) -> str:
        """String representation of DBDataset row

        Returns:
            str: String representation of DBDataset row
        """
        output = f"<Dataset(run number={self.run_number}, id={self.id}, "
        output += f"reference period={str(self.dataset_date)}, update frequency={self.update_frequency},\n"
        output += f"review date={str(self.review_date)}, last modified={str(self.last_modified)}, "
        output += f"metadata modified={str(self.metadata_modified)}, updated by script={str(self.updated_by_script)},\n"
        output += f"latest of modifieds={str(self.latest_of_modifieds)}, what updated={str(self.what_updated)},\n"
        output += f"Resource {str(self.last_resource_updated)}: last modified={str(self.last_resource_modified)},\n"
        output += f"Dataset fresh={str(self.fresh)}, error={str(self.error)})>"
        return output



"""Utility to retry HTTP requests with exponential backoff interval
"""
import asyncio
import logging
from typing import Any, Callable, List

import aiohttp
from aiohttp import ClientResponse

logger = logging.getLogger(__name__)


HTTP_STATUS_CODES_TO_RETRY = [500, 502, 503, 504]


class FailedRequest(Exception):
    """
    A wrapper for all possible exceptions during an HTTP request

    Args:
        raised (str): Exception type
        message (str): Exception message
        code (str): HTTP status code
        url (str): URL that was requested
    """

    code = 0
    message = ""
    url = ""
    raised = ""

    def __init__(
        self,
        *,
        raised: str = "",
        message: str = "",
        code: str = "",
        url: str = "",
    ):
        self.raised = raised
        self.message = message
        self.code = code
        self.url = url

        super().__init__(
            "code={c} message={m} raised={r} url={u}".format(
                c=self.code, m=self.message, r=self.raised, u=self.url
            )
        )


async def send_http(
    session: aiohttp.ClientSession,
    method: str,
    url: str,
    *,
    retries: int = 1,
    interval: int = 1,
    backoff: int = 2,
    http_status_codes_to_retry: List[int] = HTTP_STATUS_CODES_TO_RETRY,
    fn: Callable[[ClientResponse], Any] = lambda x: x,
    **kwargs: Any,
):
    """
    Send an HTTP request and implement retry logic

    Arguments:
        session (aiohttp.ClientSession): A client aiohttp session object
        method (str): Method to use eg. "get"
        url (str): URL for the request
        retries (int): Number of times to retry in case of failure
        interval (float): Time to wait before retries
        backoff (int): Multiply interval by this factor after each failure
        http_status_codes_to_retry (List[int]): List of status codes to retry
        fn (Callable[[x],x]: Function to call on successful connection
        **kwargs
    """
    backoff_interval = interval
    raised_exc = None

    if method not in ["get", "patch", "post"]:
        raise ValueError

    if retries == -1:  # -1 means retry indefinitely
        attempt = -1
    elif retries == 0:  # Zero means don't retry
        attempt = 1
    else:  # any other value means retry N times
        attempt = retries + 1

    while attempt != 0:
        if raised_exc:
            logger.error(
                f'Caught "{raised_exc}" url:{url} method:{method.upper()}, remaining tries {attempt}, '
                "sleeping {backoff_interval:.2f}secs"
            )
            await asyncio.sleep(backoff_interval)
            # bump interval for the next possible attempt
            backoff_interval *= backoff
        # logger.info(f'sending {method.upper()} {url} with {kwargs}')
        try:
            async with await getattr(session, method)(
                url, **kwargs
            ) as response:
                if response.status == 200:
                    return await fn(response)
                elif response.status in http_status_codes_to_retry:
                    logger.error(
                        f'Received invalid response code:{response.status} error:{""}'
                        f" response:{response.reason} url:{url}"
                    )
                    raise aiohttp.ClientResponseError(
                        code=response.status,
                        message=response.reason,
                        request_info=response.request_info,
                        history=response.history,
                    )
                else:
                    raise FailedRequest(
                        code=response.status,
                        message="Non-retryable response code",
                        raised="aiohttp.ClientResponseError",
                        url=url,
                    )
        except aiohttp.ClientError as exc:
            try:
                code = exc.code
            except AttributeError:
                code = ""
            raised_exc = FailedRequest(
                code=code,
                message=str(exc),
                raised=f"{exc.__class__.__module__}.{exc.__class__.__qualname__}",
                url=url,
            )
        except asyncio.TimeoutError as exc:
            raised_exc = FailedRequest(
                code="",
                message="asyncio.TimeoutError",
                raised=f"{exc.__class__.__module__}.{exc.__class__.__qualname__}",
                url=url,
            )
        else:
            raised_exc = None
            break

        attempt -= 1

    if raised_exc:
        raise raised_exc



"""SQLAlchemy class representing DBTestDate row. Holds test data for dates.
"""
from hdx.database import Base
from sqlalchemy import Column

from hdx.freshness.database import CustomDateTime


class DBTestDate(Base):
    """
    test_date = Column(CustomDateTime, primary_key=True)
    """

    test_date = Column(CustomDateTime, primary_key=True)

    def __repr__(self) -> str:
        """String representation of DBTestDate row

        Returns:
            str: String representation of DBTestDate row
        """
        return f"<Test date={str(self.test_date)}>"



"""Determine freshness for all datasets in HDX
"""
import logging
import re
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse

from dateutil.parser import ParserError
from hdx.api.configuration import Configuration
from hdx.data.dataset import Dataset
from hdx.data.hdxobject import HDXError
from hdx.data.resource import Resource
from hdx.utilities.dateparse import now_utc, parse_date
from hdx.utilities.dictandlist import (
    dict_of_lists_add,
    list_distribute_contents,
)
from sqlalchemy import and_, exists
from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm import Session

from ..database.dbdataset import DBDataset
from ..database.dbinfodataset import DBInfoDataset
from ..database.dborganization import DBOrganization
from ..database.dbresource import DBResource
from ..database.dbrun import DBRun
from ..testdata.serialize import (
    serialize_datasets,
    serialize_hashresults,
    serialize_now,
    serialize_results,
)
from ..utils.retrieval import Retrieval

logger = logging.getLogger(__name__)

default_no_urls_to_check = 1000


class DataFreshness:
    """Data freshness main class

    Args:
        session (sqlalchemy.orm.Session): Session to use for queries
        testsession (Optional[sqlalchemy.orm.Session]): Session for test data or None
        datasets (Optional[List[Dataset]]): List of datasets or read from HDX if None
        now (datetime): Date to use or take current time if None
        do_touch (bool): Whether to touch HDX resources whose hash has changed
    """

    bracketed_date = re.compile(r"\((.*)\)")

    def __init__(
        self,
        session: Session,
        testsession: Optional[Session] = None,
        datasets: Optional[List[Dataset]] = None,
        now: datetime = None,
        do_touch: bool = False,
    ) -> None:
        """"""
        self.session = session
        self.urls_to_check_count = 0
        self.updated_by_script_netlocs_checked = set()
        self.never_update = 0
        self.live_update = 0
        self.asneeded_update = 0
        self.dataset_what_updated = dict()
        self.resource_what_updated = dict()
        self.resource_last_modified_count = 0
        self.resource_broken_count = 0
        self.do_touch = do_touch

        self.url_internal = "data.humdata.org"

        self.freshness_by_frequency = dict()
        for key, value in Configuration.read()["aging"].items():
            update_frequency = int(key)
            freshness_frequency = dict()
            for status in value:
                nodays = value[status]
                freshness_frequency[status] = timedelta(days=nodays)
            self.freshness_by_frequency[update_frequency] = freshness_frequency
        self.freshness_statuses = {
            0: "0: Fresh",
            1: "1: Due",
            2: "2: Overdue",
            3: "3: Delinquent",
            None: "Freshness Unavailable",
        }
        self.testsession: Optional[Session] = testsession
        if datasets is None:  # pragma: no cover
            Configuration.read().set_read_only(
                True
            )  # so that we only get public datasets
            logger.info("Retrieving all datasets from HDX")
            self.datasets: List[Dataset] = Dataset.get_all_datasets()
            Configuration.read().set_read_only(False)
            if self.testsession:
                serialize_datasets(self.testsession, self.datasets)
        else:
            self.datasets: List[Dataset] = datasets
        if now is None:  # pragma: no cover
            self.now = now_utc()
            if self.testsession:
                serialize_now(self.testsession, self.now)
        else:
            self.now = now
        self.previous_run_number = (
            self.session.query(DBRun.run_number)
            .distinct()
            .order_by(DBRun.run_number.desc())
            .first()
        )
        if self.previous_run_number is not None:
            self.previous_run_number = self.previous_run_number[0]
            self.run_number = self.previous_run_number + 1
            no_resources = self.no_resources_force_hash()
            if no_resources:
                self.no_urls_to_check = no_resources
            else:
                self.no_urls_to_check = default_no_urls_to_check
        else:
            self.previous_run_number = None
            self.run_number = 0
            self.no_urls_to_check = default_no_urls_to_check

        logger.info(f"Will force hash {self.no_urls_to_check} resources")

    def no_resources_force_hash(self) -> Optional[int]:
        """Get number of resources to force hash

        Returns:
            Optional[int]: Number of resources to force hash or None
        """

        columns = [DBResource.id, DBDataset.updated_by_script]
        filters = [
            DBResource.dataset_id == DBDataset.id,
            DBResource.run_number == self.previous_run_number,
            DBDataset.run_number == self.previous_run_number,
            DBResource.url.notlike(f"%{self.url_internal}%"),
        ]
        query = self.session.query(*columns).filter(and_(*filters))
        noscriptupdate = 0
        noresources = 0
        for result in query:
            updated_by_script = result[1]
            if updated_by_script is not None:
                noscriptupdate += 1
                continue
            noresources += 1
        if noscriptupdate == 0:
            return None
        return noresources

    def spread_datasets(self) -> None:
        """Try to arrange the list of datasets so that downloads don't keep hitting the
        same server by moving apart datasets from the same organisation

        Returns:
            None
        """
        self.datasets: List[Dataset] = list_distribute_contents(
            self.datasets, lambda x: x["organization"]["name"]
        )

    def add_new_run(self) -> None:
        """Add a new run number with corresponding date

        Returns:
            None
        """
        dbrun = DBRun(run_number=self.run_number, run_date=self.now)
        self.session.add(dbrun)
        self.session.commit()

    @staticmethod
    def prefix_what_updated(dbresource: DBResource, prefix: str) -> None:
        """Prefix the what_updated field of resource

        Args:
            dbresource (DBResource): DBResource object to change
            prefix (str): Prefix to prepend

        Returns:
            None
        """
        what_updated = f"{prefix}-{dbresource.what_updated}"
        dbresource.what_updated = what_updated

    def process_resources(
        self,
        dataset_id: str,
        previous_dbdataset: DBDataset,
        resources: List[Resource],
        updated_by_script: Optional[datetime],
        hash_ids: List[str] = None,
    ) -> Tuple[List[Tuple], Optional[str], Optional[datetime]]:
        """Process HDX dataset's resources. If the resource has not been checked for
        30 days and we are below the threshold for resource checking, then the resource
        is flagged to be hashed even if the dataset is fresh.

        Args:
            dataset_id (str): Dataset id
            previous_dbdataset (DBDataset): DBDataset object from previous run
            resources (List[Resource]): HDX resources to process
            updated_by_script (Optional[datetime]): Time script updated or None
            hash_ids (Optional[List[str]]): Resource ids to hash for testing purposes

        Returns:
            Tuple[List[Tuple], Optional[str], Optional[datetime]]:
            (resources to download, id of last resource updated, time updated)
        """
        last_resource_updated = None
        last_resource_modified = None
        dataset_resources = list()
        for resource in resources:
            resource_id = resource["id"]
            dict_of_lists_add(self.resource_what_updated, "total", resource_id)
            url = resource["url"]
            name = resource["name"]
            metadata_modified = parse_date(
                resource["metadata_modified"], include_microseconds=True
            )
            last_modified = parse_date(
                resource["last_modified"], include_microseconds=True
            )
            if last_resource_modified:
                if last_modified > last_resource_modified:
                    last_resource_updated = resource_id
                    last_resource_modified = last_modified
            else:
                last_resource_updated = resource_id
                last_resource_modified = last_modified
            dbresource = DBResource(
                run_number=self.run_number,
                id=resource_id,
                name=name,
                dataset_id=dataset_id,
                url=url,
                last_modified=last_modified,
                metadata_modified=metadata_modified,
                latest_of_modifieds=last_modified,
                what_updated="firstrun",
            )
            if previous_dbdataset is not None:
                try:
                    previous_dbresource = (
                        self.session.query(DBResource)
                        .filter_by(
                            id=resource_id,
                            run_number=previous_dbdataset.run_number,
                        )
                        .one()
                    )
                    if last_modified > previous_dbresource.last_modified:
                        dbresource.what_updated = "filestore"
                    else:
                        dbresource.last_modified = (
                            previous_dbresource.last_modified
                        )
                        dbresource.what_updated = "nothing"
                    if (
                        last_modified
                        <= previous_dbresource.latest_of_modifieds
                    ):
                        dbresource.latest_of_modifieds = (
                            previous_dbresource.latest_of_modifieds
                        )
                    dbresource.http_last_modified = (
                        previous_dbresource.http_last_modified
                    )
                    dbresource.md5_hash = previous_dbresource.md5_hash
                    dbresource.hash_last_modified = (
                        previous_dbresource.hash_last_modified
                    )
                    dbresource.when_checked = previous_dbresource.when_checked

                except NoResultFound:
                    pass
            self.session.add(dbresource)

            should_hash = False
            if updated_by_script:
                netloc = urlparse(url).netloc
                if netloc in self.updated_by_script_netlocs_checked:
                    dict_of_lists_add(
                        self.resource_what_updated,
                        dbresource.what_updated,
                        resource_id,
                    )
                    continue
                else:
                    should_hash = True
                    self.updated_by_script_netlocs_checked.add(netloc)
            if self.url_internal in url:
                self.prefix_what_updated(dbresource, "internal")
                dict_of_lists_add(
                    self.resource_what_updated,
                    dbresource.what_updated,
                    resource_id,
                )
                continue
            if hash_ids:
                should_hash = resource_id in hash_ids
            elif not should_hash:
                should_hash = (
                    self.urls_to_check_count < self.no_urls_to_check
                    and (
                        dbresource.when_checked is None
                        or self.now - dbresource.when_checked
                        > timedelta(days=30)
                    )
                )
            resource_format = resource["format"].lower()
            dataset_resources.append(
                (
                    url,
                    resource_id,
                    resource_format,
                    dbresource.what_updated,
                    should_hash,
                )
            )
        return dataset_resources, last_resource_updated, last_resource_modified

    def process_datasets(
        self, hash_ids: Optional[List[str]] = None
    ) -> Tuple[Dict[str, str], List[Tuple]]:
        """Process HDX datasets. Extract necessary metadata and store in the
        freshness database. Calculate an initial freshness based on the metadata
        (last modified - which can change due to filestore resource changes,
        review date - when someone clicks the reviewed button the UI,
        updated by script - scripts provide the date of update in HDX metadata)
        For datasets that are not initially fresh or which have resources that have not
        been checked in the last 30 days (up to the threshold for the number of
        resources to check), the resources are flagged to be downloaded and hashed.

        Args:
            hash_ids (Optional[List[str]]): Resource ids to hash for testing purposes

        Returns:
            Tuple[Dict[str, str], List[Tuple]]: (datasets to check, resources to check)
        """
        resources_to_check = list()
        datasets_to_check = dict()
        logger.info("Processing datasets")
        for dataset in self.datasets:
            resources = dataset.get_resources()
            if dataset.is_requestable():  # ignore requestable
                continue
            dataset_id = dataset["id"]
            dict_of_lists_add(self.dataset_what_updated, "total", dataset_id)
            organization_id = dataset["organization"]["id"]
            organization_name = dataset["organization"]["name"]
            organization_title = dataset["organization"]["title"]
            try:
                dborganization = (
                    self.session.query(DBOrganization)
                    .filter_by(id=organization_id)
                    .one()
                )
                dborganization.name = organization_name
                dborganization.title = organization_title
            except NoResultFound:
                dborganization = DBOrganization(
                    name=organization_name,
                    id=organization_id,
                    title=organization_title,
                )
                self.session.add(dborganization)
            dataset_name = dataset["name"]
            dataset_title = dataset["title"]
            dataset_private = dataset["private"]
            dataset_maintainer = dataset["maintainer"]
            dataset_location = ",".join([x["name"] for x in dataset["groups"]])
            try:
                dbinfodataset = (
                    self.session.query(DBInfoDataset)
                    .filter_by(id=dataset_id)
                    .one()
                )
                dbinfodataset.name = dataset_name
                dbinfodataset.title = dataset_title
                dbinfodataset.private = dataset_private
                dbinfodataset.organization_id = organization_id
                dbinfodataset.maintainer = dataset_maintainer
                dbinfodataset.location = dataset_location
            except NoResultFound:
                dbinfodataset = DBInfoDataset(
                    name=dataset_name,
                    id=dataset_id,
                    title=dataset_title,
                    private=dataset_private,
                    organization_id=organization_id,
                    maintainer=dataset_maintainer,
                    location=dataset_location,
                )
                self.session.add(dbinfodataset)
            try:
                previous_dbdataset = (
                    self.session.query(DBDataset)
                    .filter_by(
                        run_number=self.previous_run_number, id=dataset_id
                    )
                    .one()
                )
            except NoResultFound:
                previous_dbdataset = None

            update_frequency = dataset.get("data_update_frequency")
            updated_by_script = None
            if update_frequency is not None:
                update_frequency = int(update_frequency)
                updated_by_script = dataset.get("updated_by_script")
                if updated_by_script:
                    if "freshness_ignore" in updated_by_script:
                        updated_by_script = None
                    else:
                        match = self.bracketed_date.search(updated_by_script)
                        if match is None:
                            updated_by_script = None
                        else:
                            try:
                                updated_by_script = parse_date(
                                    match.group(1), include_microseconds=True
                                )
                            except ParserError:
                                updated_by_script = None
            (
                dataset_resources,
                last_resource_updated,
                last_resource_modified,
            ) = self.process_resources(
                dataset_id,
                previous_dbdataset,
                resources,
                updated_by_script,
                hash_ids=hash_ids,
            )
            reference_period = dataset.get("dataset_date")
            metadata_modified = parse_date(
                dataset["metadata_modified"], include_microseconds=True
            )
            if "last_modified" in dataset:
                last_modified = parse_date(
                    dataset["last_modified"], include_microseconds=True
                )
            else:
                last_modified = datetime(1970, 1, 1, 0, 0, tzinfo=timezone.utc)
            if len(resources) == 0 and last_resource_updated is None:
                last_resource_updated = "NO RESOURCES"
                last_resource_modified = datetime(
                    1970, 1, 1, 0, 0, tzinfo=timezone.utc
                )
                error = True
                what_updated = "no resources"
            else:
                error = False
                what_updated = "firstrun"
            review_date = dataset.get("review_date")
            if review_date is None:
                latest_of_modifieds = last_modified
            else:
                review_date = parse_date(
                    review_date, include_microseconds=True
                )
                if review_date > last_modified:
                    latest_of_modifieds = review_date
                else:
                    latest_of_modifieds = last_modified
            if updated_by_script and updated_by_script > latest_of_modifieds:
                latest_of_modifieds = updated_by_script

            fresh = None
            if update_frequency is not None and not error:
                if update_frequency == 0:
                    fresh = 0
                    self.live_update += 1
                elif update_frequency == -1:
                    fresh = 0
                    self.never_update += 1
                elif update_frequency == -2:
                    fresh = 0
                    self.asneeded_update += 1
                else:
                    fresh = self.calculate_freshness(
                        latest_of_modifieds, update_frequency
                    )

            dbdataset = DBDataset(
                run_number=self.run_number,
                id=dataset_id,
                dataset_date=reference_period,
                update_frequency=update_frequency,
                review_date=review_date,
                last_modified=last_modified,
                metadata_modified=metadata_modified,
                updated_by_script=updated_by_script,
                latest_of_modifieds=latest_of_modifieds,
                what_updated=what_updated,
                last_resource_updated=last_resource_updated,
                last_resource_modified=last_resource_modified,
                fresh=fresh,
                error=error,
            )
            if previous_dbdataset is not None and not error:
                dbdataset.what_updated = self.add_what_updated(
                    dbdataset.what_updated, "nothing"
                )
                if (
                    last_modified > previous_dbdataset.last_modified
                ):  # filestore update would cause this
                    dbdataset.what_updated = self.add_what_updated(
                        dbdataset.what_updated, "filestore"
                    )
                else:
                    dbdataset.last_modified = previous_dbdataset.last_modified
                if previous_dbdataset.review_date is None:
                    if review_date is not None:
                        dbdataset.what_updated = self.add_what_updated(
                            dbdataset.what_updated, "review date"
                        )
                else:
                    if (
                        review_date is not None
                        and review_date > previous_dbdataset.review_date
                    ):  # someone clicked the review button
                        dbdataset.what_updated = self.add_what_updated(
                            dbdataset.what_updated, "review date"
                        )
                    else:
                        dbdataset.review_date = previous_dbdataset.review_date
                if updated_by_script and (
                    previous_dbdataset.updated_by_script is None
                    or updated_by_script > previous_dbdataset.updated_by_script
                ):  # new script update of datasets
                    dbdataset.what_updated = self.add_what_updated(
                        dbdataset.what_updated, "script update"
                    )
                else:
                    dbdataset.updated_by_script = (
                        previous_dbdataset.updated_by_script
                    )
                if (
                    last_resource_modified
                    <= previous_dbdataset.last_resource_modified
                ):
                    # we keep this so that although we don't normally use it,
                    # we retain the ability to run without touching CKAN
                    dbdataset.last_resource_updated = (
                        previous_dbdataset.last_resource_updated
                    )
                    dbdataset.last_resource_modified = (
                        previous_dbdataset.last_resource_modified
                    )
                if (
                    latest_of_modifieds
                    < previous_dbdataset.latest_of_modifieds
                ):
                    dbdataset.latest_of_modifieds = (
                        previous_dbdataset.latest_of_modifieds
                    )
                    if update_frequency is not None and update_frequency > 0:
                        fresh = self.calculate_freshness(
                            previous_dbdataset.latest_of_modifieds,
                            update_frequency,
                        )
                        dbdataset.fresh = fresh
            self.session.add(dbdataset)

            update_string = f"{self.freshness_statuses[fresh]}, Updated {dbdataset.what_updated}"
            anyresourcestohash = False
            for (
                url,
                resource_id,
                resource_format,
                what_updated,
                should_hash,
            ) in dataset_resources:
                if not should_hash:
                    if (
                        fresh == 0 and update_frequency != 1
                    ) or update_frequency is None:
                        dict_of_lists_add(
                            self.resource_what_updated,
                            what_updated,
                            resource_id,
                        )
                        continue
                resources_to_check.append(
                    (url, resource_id, resource_format, what_updated)
                )
                self.urls_to_check_count += 1
                anyresourcestohash = True
            if anyresourcestohash:
                datasets_to_check[dataset_id] = update_string
            else:
                dict_of_lists_add(
                    self.dataset_what_updated, update_string, dataset_id
                )
        self.session.commit()
        return datasets_to_check, resources_to_check

    def check_urls(
        self,
        resources_to_check: List[Tuple],
        user_agent: str,
        results: Optional[Dict] = None,
        hash_results: Optional[Dict] = None,
    ) -> Tuple[Dict[str, Tuple], Dict[str, Tuple]]:
        """Download resources and hash them. If the hash has changed compared to the
        previous run, download and hash again. Return two dictionaries, the first
        with the hashes from the first downloads and the second with the hashes from
        the second downloads.

        Args:
            resources_to_check (List[Tuple]): List of resources to be checked
            user_agent (str): User agent string to use when downloading
            results (Optional[Dict]): Test results to use in place of first downloads
            hash_results (Optional[Dict]): Test results replacing second downloads

        Returns:
            Tuple[Dict[str, Tuple], Dict[str, Tuple]]:
            (results of first download, results of second download)
        """

        def get_netloc(x):
            return urlparse(x[0]).netloc

        retrieval = Retrieval(user_agent, self.url_internal)
        if results is None:  # pragma: no cover
            resources_to_check = list_distribute_contents(
                resources_to_check, get_netloc
            )
            results = retrieval.retrieve(resources_to_check)
            if self.testsession:
                serialize_results(self.testsession, results)

        hash_check = list()
        for resource_id in results:
            (
                url,
                resource_format,
                err,
                http_last_modified,
                hash,
                xlsx_hash,
            ) = results[resource_id]
            if hash:
                dbresource = (
                    self.session.query(DBResource)
                    .filter_by(id=resource_id, run_number=self.run_number)
                    .one()
                )
                if dbresource.md5_hash == hash:  # File unchanged
                    continue
                if (
                    xlsx_hash and dbresource.md5_hash == xlsx_hash
                ):  # File unchanged
                    continue
                hash_check.append((url, resource_id, resource_format))

        if hash_results is None:  # pragma: no cover
            hash_check = list_distribute_contents(hash_check, get_netloc)
            hash_results = retrieval.retrieve(hash_check)
            if self.testsession:
                serialize_hashresults(self.testsession, hash_results)

        return results, hash_results

    def process_results(
        self,
        results: Dict[str, Tuple],
        hash_results: Dict[str, Tuple],
        resourcecls: Union[Resource, Any] = Resource,
    ) -> Dict[str, Dict[str, Tuple]]:
        """Process the downloaded and hashed resources. If the two hashes are the same
        but different to the previous run's, the file has been changed. If the two
        hashes are different, it is an API (eg. editable Google sheet) where the hash
        constantly changes. If the file is determined to have been changed, then the
        resource on HDX is touched to update its last_modified field. Return a
        dictionary of dictionaries from dataset id to resource ids to update information
        about resources including their latest_of_modifieds.

        Args:
            results (Dict[str, Tuple]): Test results to use in place of first downloads
            hash_results (Dict[str, Tuple]): Test results replacing second downloads
            resourcecls (Union[Resource, Any]): Class to use. Defaults to Resource.

        Returns:
            Dict[str, Dict[str, Tuple]]: Dataset id to resource id to resource info
        """

        def check_broken(error):
            if error == Retrieval.toolargeerror:
                return False
            if Retrieval.notmatcherror in error:
                return True
            match_error = re.search(Retrieval.clienterror_regex, error)
            if match_error:
                return True
            return False

        datasets_resourcesinfo = dict()
        for resource_id in sorted(results):
            url, _, err, http_last_modified, hash, xlsx_hash = results[
                resource_id
            ]
            dbresource = (
                self.session.query(DBResource)
                .filter_by(id=resource_id, run_number=self.run_number)
                .one()
            )
            dataset_id = dbresource.dataset_id
            resourcesinfo = datasets_resourcesinfo.get(dataset_id, dict())
            what_updated = dbresource.what_updated
            update_last_modified = False
            is_broken = False
            if http_last_modified:
                if (
                    dbresource.http_last_modified is None
                    or http_last_modified > dbresource.http_last_modified
                ):
                    dbresource.http_last_modified = http_last_modified
            if hash:
                dbresource.when_checked = self.now
                if dbresource.md5_hash == hash:  # File unchanged
                    what_updated = self.add_what_updated(
                        what_updated, "same hash"
                    )
                elif (
                    xlsx_hash and dbresource.md5_hash == xlsx_hash
                ):  # File unchanged
                    what_updated = self.add_what_updated(
                        what_updated, "same hash"
                    )
                else:  # File updated
                    hash_to_set = hash
                    (
                        hash_url,
                        _,
                        hash_err,
                        hash_http_last_modified,
                        hash_hash,
                        hash_xlsx_hash,
                    ) = hash_results[resource_id]
                    if hash_http_last_modified:
                        if (
                            dbresource.http_last_modified is None
                            or hash_http_last_modified
                            > dbresource.http_last_modified
                        ):
                            dbresource.http_last_modified = (
                                hash_http_last_modified
                            )
                    if hash_hash:
                        if hash_hash != hash:
                            if (  # Check if this is an xlsx file that has been hashed
                                hash_xlsx_hash and hash_xlsx_hash == xlsx_hash
                            ):
                                hash = xlsx_hash
                                hash_hash = hash_xlsx_hash
                                hash_to_set = hash
                        if hash_hash == hash:
                            if (
                                dbresource.md5_hash is None
                            ):  # First occurrence of resource eg. first run - don't use hash
                                # for last modified field (and hence freshness calculation)
                                dbresource.what_updated = (
                                    self.add_what_updated(
                                        what_updated, "first hash"
                                    )
                                )
                                what_updated = dbresource.what_updated
                            else:
                                # Check if hash has occurred before
                                # select distinct md5_hash from dbresources where id = '714ef7b5-a303-4e4f-be2f-03b2ce2933c7' and md5_hash='2f3cd6a6fce5ad4d7001780846ad87a7';
                                if self.session.query(
                                    exists().where(
                                        and_(
                                            DBResource.id == resource_id,
                                            DBResource.md5_hash == hash,
                                        )
                                    )
                                ).scalar():
                                    dbresource.what_updated = (
                                        self.add_what_updated(
                                            what_updated, "repeat hash"
                                        )
                                    )
                                    what_updated = dbresource.what_updated
                                else:
                                    (
                                        what_updated,
                                        _,
                                    ) = self.set_latest_of_modifieds(
                                        dbresource, self.now, "hash"
                                    )
                                    dbresource.hash_last_modified = self.now
                                    update_last_modified = True
                            dbresource.api = False
                        else:
                            hash_to_set = hash_hash
                            what_updated = self.add_what_updated(
                                what_updated, "api"
                            )
                            dbresource.api = True
                    if hash_err:
                        what_updated = self.add_what_updated(
                            what_updated, "error"
                        )
                        dbresource.error = hash_err
                        if check_broken(hash_err):
                            is_broken = True
                    dbresource.md5_hash = hash_to_set
            if err:
                dbresource.when_checked = self.now
                what_updated = self.add_what_updated(what_updated, "error")
                dbresource.error = err
                if check_broken(err):
                    is_broken = True
            resourcesinfo[resource_id] = (
                dbresource.error,
                dbresource.latest_of_modifieds,
                dbresource.what_updated,
            )
            datasets_resourcesinfo[dataset_id] = resourcesinfo
            dict_of_lists_add(
                self.resource_what_updated, what_updated, resource_id
            )
            if (
                update_last_modified and self.do_touch
            ):  # Touch resource if needed
                try:
                    logger.info(
                        f"Updating last modified for resource {resource_id}"
                    )
                    resource = resourcecls.read_from_hdx(resource_id)
                    if resource:
                        last_modified = parse_date(
                            resource["last_modified"],
                            include_microseconds=True,
                        )
                        dbdataset = (
                            self.session.query(DBDataset)
                            .filter_by(
                                id=dataset_id, run_number=self.run_number
                            )
                            .one()
                        )
                        update_frequency = dbdataset.update_frequency
                        if update_frequency > 0:
                            if (
                                self.calculate_freshness(
                                    last_modified, update_frequency
                                )
                                == 0
                            ):
                                dotouch = False
                            else:
                                dotouch = True
                        else:
                            dotouch = True
                        if dotouch:
                            dt_notz = dbresource.latest_of_modifieds.replace(
                                tzinfo=None
                            )
                            resource["last_modified"] = dt_notz.isoformat()
                            resource.update_in_hdx(
                                operation="patch",
                                batch_mode="KEEP_OLD",
                                skip_validation=True,
                                ignore_check=True,
                            )
                            self.resource_last_modified_count += 1
                            logger.info(
                                f"Resource last modified count: {self.resource_last_modified_count}"
                            )
                        else:
                            logger.info(
                                f"Didn't update last modified for resource {resource_id} as it is fresh!"
                            )
                    else:
                        logger.error(
                            f"Last modified update failed for id {resource_id}! Resource does not exist."
                        )
                except HDXError:
                    logger.exception(
                        f"Last modified update failed for id {resource_id}!"
                    )
            if is_broken and self.do_touch:
                try:
                    logger.info(f"Marking resource {resource_id} as broken")
                    resource = resourcecls.read_from_hdx(resource_id)
                    if resource:
                        resource.mark_broken()
                        self.resource_broken_count += 1
                        logger.info(
                            f"Resource broken count: {self.resource_broken_count}"
                        )
                    else:
                        logger.error(
                            f"Mark broken failed for id {resource_id}! Resource does not exist."
                        )
                except HDXError:
                    logger.exception(
                        f"Mark broken failed for id {resource_id}!"
                    )
        self.session.commit()
        return datasets_resourcesinfo

    def update_dataset_latest_of_modifieds(
        self,
        datasets_to_check: Dict[str, str],
        datasets_resourcesinfo: Dict[str, Dict[str, Tuple]],
    ) -> None:
        """Given the dictionary of dictionaries from dataset id to resource ids to
        update information about resources including their latest_of_modifieds, work
        out latest_of_modifieds for datasets and calculate freshness.

        Args:
            datasets_to_check (Dict[str, str]): Datasets with resources that were hashed
            datasets_resourcesinfo (Dict[str, Dict[str, Tuple]]): Dataset id to resource
            id to resource info

        Returns:
            None
        """

        for dataset_id in datasets_resourcesinfo:
            dbdataset = (
                self.session.query(DBDataset)
                .filter_by(id=dataset_id, run_number=self.run_number)
                .one()
            )
            dataset = datasets_resourcesinfo[dataset_id]
            dataset_latest_of_modifieds = dbdataset.latest_of_modifieds
            dataset_what_updated = dbdataset.what_updated
            last_resource_modified = dbdataset.last_resource_modified
            last_resource_updated = dbdataset.last_resource_updated
            all_errors = True
            for resource_id in sorted(dataset):
                (
                    err,
                    new_last_resource_modified,
                    new_last_resource_what_updated,
                ) = dataset[resource_id]
                if not err:
                    all_errors = False
                if new_last_resource_modified:
                    if new_last_resource_modified > last_resource_modified:
                        last_resource_updated = resource_id
                        last_resource_modified = new_last_resource_modified
                    if (
                        new_last_resource_modified
                        > dataset_latest_of_modifieds
                    ):
                        dataset_latest_of_modifieds = (
                            new_last_resource_modified
                        )
                        dataset_what_updated = new_last_resource_what_updated
            dbdataset.last_resource_updated = last_resource_updated
            dbdataset.last_resource_modified = last_resource_modified
            self.set_latest_of_modifieds(
                dbdataset, dataset_latest_of_modifieds, dataset_what_updated
            )
            update_frequency = dbdataset.update_frequency
            if update_frequency is not None and update_frequency > 0:
                dbdataset.fresh = self.calculate_freshness(
                    dbdataset.latest_of_modifieds, update_frequency
                )
            dbdataset.error = all_errors
            status = f"{self.freshness_statuses[dbdataset.fresh]}, Updated {dbdataset.what_updated}"
            if all_errors:
                status = f"{status},error"
            dict_of_lists_add(self.dataset_what_updated, status, dataset_id)
        self.session.commit()
        for dataset_id in datasets_to_check:
            if dataset_id in datasets_resourcesinfo:
                continue
            dict_of_lists_add(
                self.dataset_what_updated,
                datasets_to_check[dataset_id],
                dataset_id,
            )

    def output_counts(self) -> str:
        """Create and display output string

        Returns:
            str: Output string
        """

        def add_what_updated_str(hdxobject_what_updated):
            nonlocal output_str
            output_str += (
                f'\n* total: {len(hdxobject_what_updated["total"])} *'
            )
            for countstr in sorted(hdxobject_what_updated):
                if countstr != "total":
                    output_str += f",\n{countstr}: {len(hdxobject_what_updated[countstr])}"

        output_str = "\n*** Resources ***"
        add_what_updated_str(self.resource_what_updated)
        output_str += "\n\n*** Datasets ***"
        add_what_updated_str(self.dataset_what_updated)
        output_str += (
            f"\n\n{self.live_update} datasets have update frequency of Live"
        )
        output_str += (
            f"\n{self.never_update} datasets have update frequency of Never"
        )
        output_str += f"\n{self.asneeded_update} datasets have update frequency of As Needed"

        logger.info(output_str)
        return output_str

    @staticmethod
    def set_latest_of_modifieds(
        dbobject: Union[DBDataset, DBResource],
        modified_date: datetime,
        what_updated: str,
    ) -> Tuple[str, bool]:
        """Set latest of modifieds if provided date is greater than current and add
        to the Database object's what_updated field.

        Args:
            dbobject (Union[DBDataset, DBResource]): Database object to update
            modified_date (datetime): New modified date
            what_updated (str): What updated eg. hash

        Returns:
            Tuple[str, bool]: (DB object's what_updated, whether new date > current)
        """
        if modified_date > dbobject.latest_of_modifieds:
            dbobject.latest_of_modifieds = modified_date
            dbobject.what_updated = DataFreshness.add_what_updated(
                dbobject.what_updated, what_updated
            )
            update = True
        else:
            update = False
        return dbobject.what_updated, update

    @staticmethod
    def add_what_updated(prev_what_updated: str, what_updated: str):
        """Add to what_updated string any new cause of update (such as hash). "nothing"
        is removed if anything else is added.

        Args:
            prev_what_updated (str): Previous what_updated string
            what_updated (str): Additional what_updated string

        Returns:
            str: New what_updated string
        """
        if what_updated in prev_what_updated:
            return prev_what_updated
        if prev_what_updated != "nothing" and prev_what_updated != "firstrun":
            if what_updated != "nothing":
                return f"{prev_what_updated},{what_updated}"
            return prev_what_updated
        else:
            return what_updated

    def calculate_freshness(
        self, last_modified: datetime, update_frequency: int
    ) -> int:
        """Calculate freshness based on a last modified date and the expected update
        frequency. Returns 0 for fresh, 1 for due, 2 for overdue and 3 for delinquent.

        Args:
            last_modified (datetime): Last modified date
            update_frequency (int): Expected update frequency

        Returns:
            int: 0 for fresh, 1 for due, 2 for overdue and 3 for delinquent
        """
        delta = self.now - last_modified
        if (
            delta
            >= self.freshness_by_frequency[update_frequency]["Delinquent"]
        ):
            return 3
        elif delta >= self.freshness_by_frequency[update_frequency]["Overdue"]:
            return 2
        elif delta >= self.freshness_by_frequency[update_frequency]["Due"]:
            return 1
        return 0



"""SQLAlchemy class representing DBTestResult row. Holds test data mimicking the result
of downloading and hashing urls (first time).
"""
from hdx.database import Base
from sqlalchemy import Boolean, Column, String

from hdx.freshness.database import CustomDateTime


class DBTestResult(Base):
    """
    id = Column(String, primary_key=True)
    url = Column(String, nullable=False)
    format = Column(String, nullable=False)
    err = Column(String)
    http_last_modified = Column(CustomDateTime)
    hash = Column(String)
    xlsx_hash = Column(String)
    force_hash = Column(Boolean, nullable=False)
    """

    id = Column(String, primary_key=True)
    url = Column(String, nullable=False)
    format = Column(String, nullable=False)
    err = Column(String)
    http_last_modified = Column(CustomDateTime)
    hash = Column(String)
    xlsx_hash = Column(String)
    force_hash = Column(Boolean, nullable=False)

    def __repr__(self) -> str:
        """String representation of DBTestResult row

        Returns:
            str: String representation of DBTestResult row
        """
        output = f"<TestResult(id={self.id}, url={self.url}, "
        output += f"format={self.format}, err={self.err}\n"
        output += f"http_last_modified={str(self.http_last_modified)}, "
        output += f"hash={self.hash}, xlsx_hash={self.xlsx_hash}, "
        output += f"force_hash={str(self.force_hash)})>"
        return output



"""SQLAlchemy class representing DBTestDataset row. Holds test data for datasets with
the aim of mimicking what would come from HDX.
"""
from hdx.database import Base
from sqlalchemy import Boolean, Column, String


class DBTestDataset(Base):
    """
    id = Column(String, primary_key=True)\
    organization_id = Column(String, nullable=False)\
    organization_name = Column(String, nullable=False)\
    organization_title = Column(String, nullable=False)\
    dataset_name = Column(String, nullable=False)\
    dataset_title = Column(String, nullable=False)\
    dataset_private = Column(Boolean, nullable=False)\
    dataset_maintainer = Column(String)\
    dataset_date = Column(String)\
    update_frequency = Column(String)\
    review_date = Column(String)\
    last_modified = Column(String, nullable=False)\
    updated_by_script = Column(String)\
    metadata_modified = Column(String, nullable=False)\
    is_requestdata_type = Column(Boolean)\
    dataset_location = Column(String)
    """

    id = Column(String, primary_key=True)
    organization_id = Column(String, nullable=False)
    organization_name = Column(String, nullable=False)
    organization_title = Column(String, nullable=False)
    dataset_name = Column(String, nullable=False)
    dataset_title = Column(String, nullable=False)
    dataset_private = Column(Boolean, nullable=False)
    dataset_maintainer = Column(String)
    dataset_date = Column(String)
    update_frequency = Column(String)
    review_date = Column(String)
    last_modified = Column(String, nullable=False)
    updated_by_script = Column(String)
    metadata_modified = Column(String, nullable=False)
    is_requestdata_type = Column(Boolean)
    dataset_location = Column(String)

    def __repr__(self) -> str:
        """String representation of DBTestDataset row

        Returns:
            str: String representation of DBTestDataset row
        """
        output = f"<Dataset(id={self.id}, name={self.dataset_name}"
        output += f"organisation name={self.organization_name},\n"
        output += f"reference period={self.dataset_date}, update frequency={self.update_frequency}, "
        output += f"review_date={str(self.review_date)}, last_modified={str(self.last_modified)},"
        output += f"updated_by_script={str(self.updated_by_script)}, metadata_modified={str(self.metadata_modified)})>"
        return output



"""Functions to serialise and deserialise test data, for example for datasets
"""
from datetime import datetime
from typing import Dict, Iterable, List, Tuple

from hdx.data.dataset import Dataset
from sqlalchemy.orm import Session

from .dbtestdataset import DBTestDataset
from .dbtestdate import DBTestDate
from .dbtesthashresult import DBTestHashResult
from .dbtestresource import DBTestResource
from .dbtestresult import DBTestResult


def serialize_datasets(session: Session, datasets: List[Dataset]) -> None:
    """Serialise HDX datasets to database objects

    Args:
        session (sqlalchemy.orm.Session): Session to use for queries for test data
        datasets (List[Dataset]): HDX datasets

    Returns:
        None
    """
    for dataset in datasets:
        dataset_id = dataset["id"]
        dbtestdataset = DBTestDataset(
            id=dataset_id,
            organization_id=dataset["organization"]["id"],
            organization_name=dataset["organization"]["name"],
            organization_title=dataset["organization"]["title"],
            dataset_name=dataset["name"],
            dataset_title=dataset["title"],
            dataset_private=dataset["private"],
            dataset_maintainer=dataset["maintainer"],
            dataset_date=dataset.get("dataset_date"),
            update_frequency=dataset.get("data_update_frequency"),
            review_date=dataset["review_date"],
            last_modified=dataset["last_modified"],
            updated_by_script=dataset.get("updated_by_script"),
            metadata_modified=dataset["metadata_modified"],
            is_requestdata_type=dataset.get("is_requestdata_type"),
            dataset_location=",".join([x["name"] for x in dataset["groups"]]),
        )
        session.add(dbtestdataset)
        for resource in dataset.get_resources():
            dbtestresource = DBTestResource(
                id=resource["id"],
                name=resource["name"],
                dataset_id=dataset_id,
                format=resource["format"],
                url=resource["url"],
                metadata_modified=resource["metadata_modified"],
                last_modified=resource["last_modified"],
            )
            session.add(dbtestresource)
    session.commit()


def serialize_now(session: Session, now: datetime) -> None:
    """Serialise date to database object

    Args:
        session (sqlalchemy.orm.Session): Session to use for queries for test data
        now (datetime): Current date

    Returns:
        None
    """
    dbtestdate = DBTestDate(test_date=now)
    session.add(dbtestdate)
    session.commit()


def serialize_results(session: Session, results: Dict[str, Tuple]) -> None:
    """Serialise results of downloading and hashing urls (first time) to database
    objects

    Args:
        session (sqlalchemy.orm.Session): Session to use for queries for test data
        results (Dict[str, Tuple]): Results of downloading and hashing urls (1st time)

    Returns:
        None
    """
    for id in results:
        (
            url,
            resource_format,
            err,
            http_last_modified,
            hash,
            xlsx_hash,
            force_hash,
        ) = results[id]
        dbtestresult = DBTestResult(
            id=id,
            url=url,
            format=resource_format,
            err=err,
            http_last_modified=http_last_modified,
            hash=hash,
            xlsx_hash=xlsx_hash,
            force_hash=force_hash,
        )
        session.add(dbtestresult)
    session.commit()


def serialize_hashresults(
    session: Session, hash_results: Dict[str, Tuple]
) -> None:
    """Serialise results of downloading and hashing urls (second time) to database
    objects

    Args:
        session (sqlalchemy.orm.Session): Session to use for queries for test data
        hash_results (Dict[str, Tuple]): Results of downloading+hashing urls (2nd time)

    Returns:
        None
    """
    for id in hash_results:
        (
            url,
            resource_format,
            err,
            http_last_modified,
            hash,
            xlsx_hash,
            force_hash,
        ) = hash_results[id]
        dbtesthashresult = DBTestHashResult(
            id=id,
            url=url,
            format=resource_format,
            err=err,
            http_last_modified=http_last_modified,
            hash=hash,
            xlsx_hash=xlsx_hash,
            force_hash=force_hash,
        )
        session.add(dbtesthashresult)
    session.commit()


def deserialize_datasets(session: Session) -> Iterable[Dataset]:
    """Deserialise database objects to HDX datasets

    Args:
        session (sqlalchemy.orm.Session): Session to use for queries for test data

    Returns:
        Iterable[Dataset]: HDX Dataset objects
    """
    datasets = dict()
    for dbtestdataset in session.query(DBTestDataset):
        dataset_id = dbtestdataset.id
        organization = {
            "id": dbtestdataset.organization_id,
            "name": dbtestdataset.organization_name,
            "title": dbtestdataset.organization_title,
        }
        dataset = Dataset(
            {
                "id": dataset_id,
                "organization": organization,
                "name": dbtestdataset.dataset_name,
                "title": dbtestdataset.dataset_title,
                "private": dbtestdataset.dataset_private,
                "maintainer": dbtestdataset.dataset_maintainer,
                "dataset_date": dbtestdataset.dataset_date,
                "data_update_frequency": dbtestdataset.update_frequency,
                "review_date": dbtestdataset.review_date,
                "last_modified": dbtestdataset.last_modified,
                "updated_by_script": dbtestdataset.updated_by_script,
                "metadata_modified": dbtestdataset.metadata_modified,
                "groups": [
                    {"name": x}
                    for x in dbtestdataset.dataset_location.split(",")
                ],
            }
        )
        dataset.set_requestable(dbtestdataset.is_requestdata_type)
        datasets[dataset_id] = dataset
    for dbtestresource in session.query(DBTestResource):
        dataset = datasets[dbtestresource.dataset_id]
        resource = {
            "id": dbtestresource.id,
            "name": dbtestresource.name,
            "format": dbtestresource.format,
            "url": dbtestresource.url,
            "metadata_modified": dbtestresource.metadata_modified,
            "last_modified": dbtestresource.last_modified,
        }
        dataset.get_resources().append(resource)
    return datasets.values()


def deserialize_now(session: Session) -> datetime:
    """Deserialise database object to date

    Args:
        session (sqlalchemy.orm.Session): Session to use for queries for test data

    Returns:
        datetime: date
    """
    return session.query(DBTestDate.test_date).scalar()


def deserialize_results(session: Session) -> List[Tuple]:
    """Deserialise database objects to results of downloading and hashing urls
    (first time)

    Args:
        session (sqlalchemy.orm.Session): Session to use for queries for test data

    Returns:
        List[Tuple]: Results of downloading and hashing urls (first time)
    """
    results = dict()
    for dbtestresult in session.query(DBTestResult):
        results[dbtestresult.id] = (
            dbtestresult.url,
            dbtestresult.format,
            dbtestresult.err,
            dbtestresult.http_last_modified,
            dbtestresult.hash,
            dbtestresult.xlsx_hash,
        )
    return results


def deserialize_hashresults(session: Session) -> List[Tuple]:
    """Deserialise database objects to results of downloading and hashing urls
    (second time)

    Args:
        session (sqlalchemy.orm.Session): Session to use for queries for test data

    Returns:
        List[Tuple]: Results of downloading and hashing urls (second time)
    """
    hash_results = dict()
    for dbtesthashresult in session.query(DBTestHashResult):
        hash_results[dbtesthashresult.id] = (
            dbtesthashresult.url,
            dbtesthashresult.format,
            dbtesthashresult.err,
            dbtesthashresult.http_last_modified,
            dbtesthashresult.hash,
            dbtesthashresult.xlsx_hash,
        )
    return hash_results



"""
Unit tests for the retrieval class.

"""
import re
from datetime import datetime, timezone

from hdx.freshness.utils.retrieval import Retrieval


class TestRetrieve:
    def test_retrieve(self):
        url1 = "https://www.rfc-editor.org/rfc/rfc7235.txt"
        url2 = "https://github.com/mcarans/hdx-data-freshness/raw/d1616d76c3b6b8ef5029eb6964b93cde688efd53/tests/fixtures/day0/notfound"
        url3 = "file://lala:10"
        url4 = "https://ocha-dap.github.io/hdx-data-freshness/tests/fixtures/retrieve/COD_MOZ_Admin0.geojson"
        url5 = "https://ocha-dap.github.io/hdx-data-freshness/tests/fixtures/retrieve/hotosm_nic_airports_lines_shp.zip"
        url6 = "https://ocha-dap.github.io/hdx-data-freshness/tests/fixtures/retrieve/Dataset.csv"
        url7 = "https://ocha-dap.github.io/hdx-data-freshness/tests/fixtures/retrieve/list_one.xls"
        url8 = "https://ocha-dap.github.io/hdx-data-freshness/tests/fixtures/retrieve/ACLED-Country-Coverage-and-ISO-Codes_8.2019.xlsx"
        url9 = "https://ocha-dap.github.io/hdx-data-freshness/tests/fixtures/retrieve/response.html"
        urls = [
            (url1, "1", "html"),
            (url2, "2", "csv"),
            (url3, "3", "csv"),
            (url4, "4", "geojson"),
            (url5, "5", "shp"),
            (url6, "6", "csv"),
            (url7, "7", "xls"),
            (url8, "8", "xlsx"),
            (url9, "9", "html"),
            (url9, "10", "csv"),
            (url9, "11", "xls"),
        ]
        result = Retrieval("test", url_ignore="data.humdata.org").retrieve(
            urls
        )
        assert result["1"] == (
            url1,
            "html",
            None,
            datetime(2014, 6, 7, 0, 53, 56, tzinfo=timezone.utc),
            "c9ef27e27f0bb5a772e12af426102804",
            None,
        )
        assert result["2"] == (
            url2,
            "csv",
            "code=404 message=Non-retryable response code raised=aiohttp.ClientResponseError url=https://github.com/mcarans/hdx-data-freshness/raw/d1616d76c3b6b8ef5029eb6964b93cde688efd53/tests/fixtures/day0/notfound",
            None,
            None,
            None,
        )
        assert result["3"][0] == url3
        regexp = r"^code= message=Cannot connect to host lala:10 ssl:default \[.*\] raised=aiohttp\.client_exceptions\.ClientConnectorError url=file:\/\/lala:10$"
        assert re.search(regexp, result["3"][2]) is not None
        assert result["3"][3] is None
        assert result["3"][4] is None
        assert result["4"][0] == url4
        assert result["4"][1] == "geojson"
        assert result["4"][4] == "c69a7ea486af00be2c409ef3b7894514"
        assert result["5"][0] == url5
        assert result["5"][1] == "shp"
        assert result["5"][4] == "681ac397411c62c3aba4c630a2fd27da"
        assert result["6"][0] == url6
        assert result["6"][1] == "csv"
        assert result["6"][4] == "50cd43aeecf3742f7fb3b8a7fb18af89"
        assert result["7"][0] == url7
        assert result["7"][1] == "xls"
        assert result["7"][4] == "d5177e2dadba53a36ae2323037e06c96"
        assert result["8"][0] == url8
        assert result["8"][1] == "xlsx"
        assert result["8"][4] == "74f72b149defd3f1a3c9000600734a96"
        assert result["8"][5] == "c3d51c5b077a48221e77797f7e771d1f"
        assert result["9"][0] == url9
        assert result["9"][1] == "html"
        assert result["9"][4] == "9070c22e0c8c41204c6d4e509929246e"
        assert result["10"][0] == url9
        assert result["10"][1] == "csv"
        assert (
            result["10"][2]
            == "File mimetype text/html; charset=utf-8 does not match HDX format csv!"
        )
        assert result["10"][4] == "9070c22e0c8c41204c6d4e509929246e"
        assert result["11"][0] == url9
        assert result["11"][1] == "xls"
        assert (
            result["11"][2]
            == "File mimetype text/html; charset=utf-8 does not match HDX format xls! File signature b'<htm' does not match HDX format xls!"
        )
        assert result["11"][4] == "9070c22e0c8c41204c6d4e509929246e"



"""
Unit tests for the freshness class.

Takes base database from fixtures/day0/test_freshness.db

Run data comes from fixtures/dayN/test_serialize.db

When adding new test data, remember that self.now (run date) is 2017-12-19 10:53:28.606889

Set force_hash in dbtestresults and dbtesthashresults if dataset is fresh or you'll get duplicates
(number of resources won't sum to total)!
"""
import os
import shutil
from os.path import join

import pytest
from hdx.database import Database
from hdx.utilities.dateparse import parse_date

from hdx.freshness.app.datafreshness import DataFreshness
from hdx.freshness.database.dbdataset import DBDataset
from hdx.freshness.database.dbinfodataset import DBInfoDataset
from hdx.freshness.database.dborganization import DBOrganization
from hdx.freshness.database.dbresource import DBResource
from hdx.freshness.database.dbrun import DBRun
from hdx.freshness.testdata.dbtestresult import DBTestResult
from hdx.freshness.testdata.serialize import (
    deserialize_datasets,
    deserialize_hashresults,
    deserialize_now,
    deserialize_results,
)


class TestFreshnessDayN:
    @pytest.fixture(scope="function")
    def database(self):
        dbfile = "test_freshness.db"
        dbpath = join("tests", dbfile)
        try:
            os.remove(dbpath)
        except FileNotFoundError:
            pass
        shutil.copyfile(join("tests", "fixtures", "day0", dbfile), dbpath)
        return {"dialect": "sqlite", "database": dbpath}

    @pytest.fixture(scope="class")
    def serializedbsession(self):
        dbpath = join("tests", "fixtures", "dayN", "test_serialize.db")
        return Database.get_session(f"sqlite:///{dbpath}")

    @pytest.fixture(scope="function")
    def now(self, serializedbsession):
        return deserialize_now(serializedbsession)

    @pytest.fixture(scope="function")
    def datasets(self, serializedbsession):
        return deserialize_datasets(serializedbsession)

    @pytest.fixture(scope="function")
    def results(self, serializedbsession):
        return deserialize_results(serializedbsession)

    @pytest.fixture(scope="function")
    def hash_results(self, serializedbsession):
        return deserialize_hashresults(serializedbsession)

    @pytest.fixture(scope="function")
    def forced_hash_ids(self, serializedbsession):
        forced_hash_ids = serializedbsession.query(DBTestResult.id).filter_by(
            force_hash=1
        )
        return [x[0] for x in forced_hash_ids]

    def test_generate_dataset(
        self,
        configuration,
        database,
        now,
        datasets,
        results,
        hash_results,
        forced_hash_ids,
        resourcecls,
    ):
        with Database(**database) as session:
            freshness = DataFreshness(
                session=session, datasets=datasets, now=now, do_touch=True
            )
            freshness.spread_datasets()
            freshness.add_new_run()
            dbsession = freshness.session
            # insert resource with run number -1 with hash 999 to test repeated hash
            modified = parse_date(
                "2015-05-07 14:44:56.599079", include_microseconds=True
            )
            hash_modified = parse_date(
                "2017-12-16 16:03:33.208327", include_microseconds=True
            )
            dbresource = DBResource(
                run_number=-1,
                id="010ab2d2-8f98-409b-a1f0-4707ad6c040a",
                name="sidih_190.csv",
                dataset_id="54d6b4b8-8cc9-42d3-82ce-3fa4fd3d9be1",
                url="https://ds-ec2.scraperwiki.com/egzfk1p/siqsxsgjnxgk3r2/cgi-bin/csv/sidih_190.csv",
                last_modified=modified,
                metadata_modified=modified,
                latest_of_modifieds=modified,
                what_updated="",
                http_last_modified=None,
                md5_hash="999",
                hash_last_modified=hash_modified,
                when_checked=hash_modified,
                api=False,
                error=None,
            )
            dbsession.add(dbresource)
            datasets_to_check, resources_to_check = freshness.process_datasets(
                hash_ids=forced_hash_ids
            )
            results, hash_results = freshness.check_urls(
                resources_to_check,
                "test",
                results=results,
                hash_results=hash_results,
            )
            resourcecls.populate_resourcedict(datasets)
            datasets_lastmodified = freshness.process_results(
                results, hash_results, resourcecls=resourcecls
            )
            freshness.update_dataset_latest_of_modifieds(
                datasets_to_check, datasets_lastmodified
            )
            output = freshness.output_counts()
            # Make sure the sum of the resources = the total resources and sum of the datasets = the total datasets!
            assert (
                output
                == """
*** Resources ***
* total: 657 *,
api: 3,
error: 26,
first hash: 4,
hash: 3,
internal-filestore: 10,
internal-nothing: 46,
nothing: 558,
repeat hash: 1,
same hash: 6

*** Datasets ***
* total: 104 *,
0: Fresh, Updated filestore: 4,
0: Fresh, Updated filestore,review date: 1,
0: Fresh, Updated firstrun: 1,
0: Fresh, Updated hash: 3,
0: Fresh, Updated nothing: 59,
0: Fresh, Updated review date: 1,
0: Fresh, Updated script update: 1,
1: Due, Updated nothing: 1,
2: Overdue, Updated nothing: 1,
3: Delinquent, Updated nothing: 23,
3: Delinquent, Updated nothing,error: 4,
Freshness Unavailable, Updated no resources: 1,
Freshness Unavailable, Updated nothing: 4

15 datasets have update frequency of Live
19 datasets have update frequency of Never
0 datasets have update frequency of As Needed"""
            )

            dbrun = dbsession.query(DBRun).filter_by(run_number=1).one()
            assert (
                str(dbrun)
                == "<Run number=1, Run date=2017-12-19 10:53:28.606889+00:00>"
            )
            dbresource = dbsession.query(DBResource).first()
            assert (
                str(dbresource)
                == """<Resource(run number=0, id=b21d6004-06b5-41e5-8e3e-0f28140bff64, name=Topline Numbers.csv, dataset id=a2150ad9-2b87-49f5-a6b2-c85dff366b75,
url=https://docs.google.com/spreadsheets/d/e/2PACX-1vRjFRZGLB8IMp0anSGR1tcGxwJgkyx0bTN9PsinqtaLWKHBEfz77LkinXeVqIE_TsGVt-xM6DQzXpkJ/pub?gid=0&single=true&output=csv,
last modified=2017-12-16 15:11:15.202742+00:00, metadata modified=2017-12-16 15:11:15.202742+00:00,
latest of modifieds=2017-12-16 15:11:15.202742+00:00, what updated=first hash,
http last modified=None,
MD5 hash=be5802368e5a6f7ad172f27732001f3a, hash last modified=None, when checked=2017-12-18 16:03:33.208327+00:00,
api=False, error=None)>"""
            )
            dbresource = (
                dbsession.query(DBResource)
                .filter(
                    DBResource.run_number == 1,
                    DBResource.id == "7b82976a-ae81-4cef-a76f-12ba14152086",
                )
                .first()
            )
            assert (
                str(dbresource)
                == """<Resource(run number=1, id=7b82976a-ae81-4cef-a76f-12ba14152086, name=Guinea, Liberia, Mali and Sierra Leone Health Facilities, dataset id=ce876595-1263-4df6-a8ca-459f92c532e4,
url=https://docs.google.com/a/megginson.com/spreadsheets/d/1paoIpHiYo7dy_dnf_luUSfowWDwNAWwS3z4GHL2J7Rc/export?format=xlsx&id=1paoIpHiYo7dy_dnf_luUSfowWDwNAWwS3z4GHL2J7Rc,
last modified=2017-12-18 22:21:26.783801+00:00, metadata modified=2017-12-18 22:21:26.783801+00:00,
latest of modifieds=2017-12-19 10:53:28.606889+00:00, what updated=hash,
http last modified=None,
MD5 hash=789, hash last modified=2017-12-19 10:53:28.606889+00:00, when checked=2017-12-19 10:53:28.606889+00:00,
api=False, error=None)>"""
            )
            count = (
                dbsession.query(DBResource)
                .filter(DBResource.url.like("%data.humdata.org%"))
                .count()
            )
            assert count == 112
            count = (
                dbsession.query(DBResource)
                .filter_by(run_number=1, what_updated="filestore", error=None)
                .count()
            )
            assert count == 0
            count = (
                dbsession.query(DBResource)
                .filter_by(run_number=1, what_updated="first hash", error=None)
                .count()
            )
            assert count == 4
            count = (
                dbsession.query(DBResource)
                .filter_by(run_number=1, what_updated="hash", error=None)
                .count()
            )
            assert count == 3
            count = (
                dbsession.query(DBResource)
                .filter_by(
                    run_number=1, what_updated="http header", error=None
                )
                .count()
            )
            assert count == 0
            count = (
                dbsession.query(DBResource)
                .filter_by(run_number=1, api=True)
                .count()
            )
            assert count == 3
            count = (
                dbsession.query(DBResource)
                .filter_by(run_number=1, what_updated="internal-nothing")
                .filter(DBResource.error.isnot(None))
                .count()
            )
            assert count == 0
            # select what_updated, api from dbresources where run_number=0 and md5_hash is not null and id in (select id from dbresources where run_number=1 and what_updated like '%hash%');
            hash_updated = (
                dbsession.query(DBResource.id)
                .filter_by(run_number=1)
                .filter(DBResource.what_updated.like("%hash%"))
            )
            assert hash_updated.count() == 8
            count = (
                dbsession.query(DBResource)
                .filter_by(run_number=0)
                .filter(DBResource.md5_hash.isnot(None))
                .filter(DBResource.id.in_(hash_updated.as_scalar()))
                .count()
            )
            assert count == 4
            dbdataset = dbsession.query(DBDataset).first()
            assert (
                str(dbdataset)
                == """<Dataset(run number=0, id=a2150ad9-2b87-49f5-a6b2-c85dff366b75, reference period=09/21/2017, update frequency=1,
review date=None, last modified=2017-12-16 15:11:15.204215+00:00, metadata modified=2017-12-16 15:11:15.204215+00:00, updated by script=None,
latest of modifieds=2017-12-16 15:11:15.204215+00:00, what updated=firstrun,
Resource b21d6004-06b5-41e5-8e3e-0f28140bff64: last modified=2017-12-16 15:11:15.202742+00:00,
Dataset fresh=2, error=False)>"""
            )
            count = (
                dbsession.query(DBDataset)
                .filter_by(run_number=1, fresh=0, what_updated="filestore")
                .count()
            )
            assert count == 4
            count = (
                dbsession.query(DBDataset)
                .filter_by(
                    run_number=1, fresh=0, what_updated="nothing", error=False
                )
                .count()
            )
            assert count == 59
            count = (
                dbsession.query(DBDataset)
                .filter_by(run_number=1, fresh=0, what_updated="review date")
                .count()
            )
            assert count == 1
            count = (
                dbsession.query(DBDataset)
                .filter_by(run_number=1, fresh=0, what_updated="script update")
                .count()
            )
            assert count == 1
            count = (
                dbsession.query(DBDataset)
                .filter_by(run_number=1, fresh=1, what_updated="nothing")
                .count()
            )
            assert count == 1
            count = (
                dbsession.query(DBDataset)
                .filter_by(
                    run_number=1, fresh=2, what_updated="nothing", error=False
                )
                .count()
            )
            assert count == 1
            count = (
                dbsession.query(DBDataset)
                .filter_by(
                    run_number=1, fresh=3, what_updated="nothing", error=False
                )
                .count()
            )
            assert count == 23
            count = (
                dbsession.query(DBDataset)
                .filter_by(
                    run_number=1, fresh=3, what_updated="nothing", error=True
                )
                .count()
            )
            assert count == 4
            count = (
                dbsession.query(DBDataset)
                .filter_by(
                    run_number=1,
                    fresh=None,
                    what_updated="nothing",
                    error=False,
                )
                .count()
            )
            assert count == 4
            count = (
                dbsession.query(DBDataset)
                .filter_by(
                    run_number=1,
                    fresh=None,
                    what_updated="nothing",
                    error=True,
                )
                .count()
            )
            assert count == 0
            count = (
                dbsession.query(DBDataset)
                .filter_by(
                    run_number=1,
                    fresh=None,
                    what_updated="no resources",
                    error=True,
                )
                .count()
            )
            assert count == 1
            dbinfodataset = dbsession.query(DBInfoDataset).first()
            assert (
                str(dbinfodataset)
                == """<InfoDataset(id=a2150ad9-2b87-49f5-a6b2-c85dff366b75, name=rohingya-displacement-topline-figures, title=Rohingya Displacement Topline Figures,
private=False, organization id=hdx,
maintainer=7d7f5f8d-7e3b-483a-8de1-2b122010c1eb, location=bgd)>"""
            )
            count = dbsession.query(DBInfoDataset).count()
            assert count == 104
            dborganization = dbsession.query(DBOrganization).first()
            assert (
                str(dborganization)
                == """<Organization(id=hdx, name=hdx, title=HDX)>"""
            )
            count = dbsession.query(DBOrganization).count()
            assert count == 40

            assert freshness.resource_last_modified_count == 2

            freshness.previous_run_number = freshness.run_number
            assert freshness.no_resources_force_hash() == 600



"""Global fixtures"""
from collections import UserDict
from os.path import join

import pytest
from hdx.api.configuration import Configuration


@pytest.fixture(scope="session")
def configuration():
    project_config_yaml = join(
        "src", "hdx", "freshness", "app", "project_configuration.yml"
    )
    Configuration._create(
        hdx_site="prod",
        user_agent="test",
        hdx_read_only=True,
        project_config_yaml=project_config_yaml,
    )


@pytest.fixture(scope="session")
def resourcecls():
    class MyResource(UserDict):
        touched = False
        broken = False
        resourcedict = None

        def __init__(self, id):
            self.data = self.resourcedict[id]

        @classmethod
        def populate_resourcedict(cls, datasets):
            cls.resourcedict = dict()
            for dataset in datasets:
                for resource in dataset.get_resources():
                    cls.resourcedict[resource["id"]] = resource

        @staticmethod
        def read_from_hdx(id):
            return MyResource(id)

        @classmethod
        def update_in_hdx(
            cls, operation, batch_mode, skip_validation, ignore_check
        ):
            cls.touched = True

        @classmethod
        def mark_broken(cls):
            cls.broken = True

    return MyResource



from datetime import datetime, timezone
from typing import Optional

from sqlalchemy import DateTime, TypeDecorator


class CustomDateTime(TypeDecorator):
    impl = DateTime
    cache_ok = True

    def process_bind_param(self, value: Optional[datetime], dialect):
        if value is None:
            return value
        if value.tzinfo:
            value = value.astimezone(timezone.utc)

        return value.replace(tzinfo=None)

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        if value.tzinfo is None:
            return value.replace(tzinfo=timezone.utc)

        return value.astimezone(timezone.utc)



"""Entry point to start data freshness
"""
import argparse
import logging
from os import getenv
from typing import Optional

from hdx.database import Database
from hdx.database.dburi import get_params_from_connection_uri
from hdx.facades.keyword_arguments import facade
from hdx.utilities.dictandlist import args_to_dict
from hdx.utilities.easy_logging import setup_logging
from hdx.utilities.path import script_dir_plus_file
from hdx.utilities.useragent import UserAgent

from hdx.freshness.app import __version__
from hdx.freshness.app.datafreshness import DataFreshness

setup_logging()
logger = logging.getLogger(__name__)


def main(
    db_uri: Optional[str] = None,
    db_params: Optional[str] = None,
    do_touch: bool = True,
    save: bool = False,
    **ignore,
) -> None:
    """Run freshness. Either a database connection string (db_uri) or database
    connection parameters (db_params) can be supplied. If neither is supplied, a local
    SQLite database with filename "freshness.db" is assumed.

    Args:
        db_uri (Optional[str]): Database connection URI. Defaults to None.
        db_params (Optional[str]): Database connection parameters. Defaults to None.
        do_touch (bool): Touch HDX datasets if files change. Defaults to False.
        save (bool): Whether to save state for testing. Defaults to False.

    Returns:
        None
    """
    logger.info(f"> Data freshness {__version__}")
    if db_params:
        params = args_to_dict(db_params)
    elif db_uri:
        params = get_params_from_connection_uri(db_uri)
    else:
        params = {"dialect": "sqlite", "database": "freshness.db"}
    logger.info(f"> Database parameters: {params}")
    with Database(**params) as session:
        testsession = None
        if save:
            testsession = Database.get_session("sqlite:///test_serialize.db")
        # Setup including reading all datasets from HDX and setting threshold for how
        # many resources to force hash
        freshness = DataFreshness(
            session=session, testsession=testsession, do_touch=do_touch
        )
        # Arrange order of list of datasets so that datasets from the same organisation
        # are moved away from each other
        freshness.spread_datasets()
        # Add new run number and date
        freshness.add_new_run()
        # Store metadata in freshness database, calculate an initial freshness based on
        # the metadata and determine resources to be downloaded and hashed
        datasets_to_check, resources_to_check = freshness.process_datasets()
        # Download resource urls and hash them
        results, hash_results = freshness.check_urls(
            resources_to_check, UserAgent.get()
        )
        # Work out if hashed files have changed and if so, touch resources.
        # Determine latest of modifieds and freshness for all datasets
        datasets_resourcesinfo = freshness.process_results(
            results, hash_results
        )
        # Work out latest_of_modifieds for datasets and calculate freshness
        freshness.update_dataset_latest_of_modifieds(
            datasets_to_check, datasets_resourcesinfo
        )
        # Display output string
        freshness.output_counts()
        if testsession:
            testsession.close()
    logger.info("Freshness completed!")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Data Freshness")
    parser.add_argument("-hk", "--hdx_key", default=None, help="HDX api key")
    parser.add_argument("-ua", "--user_agent", default=None, help="user agent")
    parser.add_argument("-pp", "--preprefix", default=None, help="preprefix")
    parser.add_argument(
        "-hs", "--hdx_site", default=None, help="HDX site to use"
    )
    parser.add_argument(
        "-db", "--db_uri", default=None, help="Database connection string"
    )
    parser.add_argument(
        "-dp",
        "--db_params",
        default=None,
        help="Database connection parameters. Overrides --db_uri.",
    )
    parser.add_argument(
        "-dt",
        "--donttouch",
        default=False,
        action="store_true",
        help="Don't touch datasets",
    )
    parser.add_argument(
        "-s",
        "--save",
        default=False,
        action="store_true",
        help="Save state for testing",
    )
    args = parser.parse_args()
    hdx_key = args.hdx_key
    if hdx_key is None:
        hdx_key = getenv("HDX_KEY")
    user_agent = args.user_agent
    if user_agent is None:
        user_agent = getenv("USER_AGENT")
        if user_agent is None:
            user_agent = "freshness"
    preprefix = args.preprefix
    if preprefix is None:
        preprefix = getenv("PREPREFIX")
    hdx_site = args.hdx_site
    if hdx_site is None:
        hdx_site = getenv("HDX_SITE", "prod")
    db_uri = args.db_uri
    if db_uri is None:
        db_uri = getenv("DB_URI")
    if db_uri and "://" not in db_uri:
        db_uri = f"postgresql://{db_uri}"
    project_config_yaml = script_dir_plus_file(
        "project_configuration.yml", main
    )
    facade(
        main,
        hdx_key=hdx_key,
        user_agent=user_agent,
        preprefix=preprefix,
        hdx_site=hdx_site,
        project_config_yaml=project_config_yaml,
        db_uri=db_uri,
        db_params=args.db_params,
        do_touch=not args.donttouch,
        save=args.save,
    )



"""
Unit tests for the freshness class.

"""
import os
from os.path import join

import pytest
from hdx.database import Database

from hdx.freshness.app.datafreshness import DataFreshness
from hdx.freshness.database.dbdataset import DBDataset
from hdx.freshness.database.dbinfodataset import DBInfoDataset
from hdx.freshness.database.dborganization import DBOrganization
from hdx.freshness.database.dbresource import DBResource
from hdx.freshness.database.dbrun import DBRun
from hdx.freshness.testdata.dbtestresult import DBTestResult
from hdx.freshness.testdata.serialize import (
    deserialize_datasets,
    deserialize_hashresults,
    deserialize_now,
    deserialize_results,
)


class TestFreshnessDay0:
    @pytest.fixture(scope="function")
    def nodatabase(self):
        dbpath = join("tests", "test_freshness.db")
        try:
            os.remove(dbpath)
        except FileNotFoundError:
            pass
        return {"dialect": "sqlite", "database": dbpath}

    @pytest.fixture(scope="class")
    def serializedbsession(self):
        dbpath = join("tests", "fixtures", "day0", "test_serialize.db")
        return Database.get_session(f"sqlite:///{dbpath}")

    @pytest.fixture(scope="function")
    def now(self, serializedbsession):
        return deserialize_now(serializedbsession)

    @pytest.fixture(scope="function")
    def datasets(self, serializedbsession):
        return deserialize_datasets(serializedbsession)

    @pytest.fixture(scope="function")
    def results(self, serializedbsession):
        return deserialize_results(serializedbsession)

    @pytest.fixture(scope="function")
    def hash_results(self, serializedbsession):
        return deserialize_hashresults(serializedbsession)

    @pytest.fixture(scope="function")
    def forced_hash_ids(self, serializedbsession):
        forced_hash_ids = serializedbsession.query(DBTestResult.id).filter_by(
            force_hash=1
        )
        return [x[0] for x in forced_hash_ids]

    def test_generate_dataset(
        self,
        configuration,
        nodatabase,
        now,
        datasets,
        results,
        hash_results,
        forced_hash_ids,
        resourcecls,
    ):
        with Database(**nodatabase) as session:
            freshness = DataFreshness(
                session=session, datasets=datasets, now=now
            )
            freshness.spread_datasets()
            freshness.add_new_run()
            datasets_to_check, resources_to_check = freshness.process_datasets(
                hash_ids=forced_hash_ids
            )
            results, hash_results = freshness.check_urls(
                resources_to_check,
                "test",
                results=results,
                hash_results=hash_results,
            )
            resourcecls.populate_resourcedict(datasets)
            datasets_lastmodified = freshness.process_results(
                results, hash_results, resourcecls=resourcecls
            )
            freshness.update_dataset_latest_of_modifieds(
                datasets_to_check, datasets_lastmodified
            )
            output = freshness.output_counts()
            assert (
                output
                == """
*** Resources ***
* total: 660 *,
api: 4,
error: 27,
first hash: 9,
firstrun: 564,
internal-firstrun: 56

*** Datasets ***
* total: 103 *,
0: Fresh, Updated firstrun: 67,
1: Due, Updated firstrun: 1,
2: Overdue, Updated firstrun: 1,
3: Delinquent, Updated firstrun: 25,
3: Delinquent, Updated firstrun,error: 5,
Freshness Unavailable, Updated firstrun: 4

15 datasets have update frequency of Live
19 datasets have update frequency of Never
0 datasets have update frequency of As Needed"""
            )

            dbsession = freshness.session
            dbrun = dbsession.query(DBRun).one()
            assert (
                str(dbrun)
                == "<Run number=0, Run date=2017-12-18 16:03:33.208327+00:00>"
            )
            dbresource = dbsession.query(DBResource).first()
            assert (
                str(dbresource)
                == """<Resource(run number=0, id=b21d6004-06b5-41e5-8e3e-0f28140bff64, name=Topline Numbers.csv, dataset id=a2150ad9-2b87-49f5-a6b2-c85dff366b75,
url=https://docs.google.com/spreadsheets/d/e/2PACX-1vRjFRZGLB8IMp0anSGR1tcGxwJgkyx0bTN9PsinqtaLWKHBEfz77LkinXeVqIE_TsGVt-xM6DQzXpkJ/pub?gid=0&single=true&output=csv,
last modified=2017-12-16 15:11:15.202742+00:00, metadata modified=2017-12-16 15:11:15.202742+00:00,
latest of modifieds=2017-12-16 15:11:15.202742+00:00, what updated=first hash,
http last modified=None,
MD5 hash=be5802368e5a6f7ad172f27732001f3a, hash last modified=None, when checked=2017-12-18 16:03:33.208327+00:00,
api=False, error=None)>"""
            )
            count = (
                dbsession.query(DBResource)
                .filter(DBResource.url.like("%data.humdata.org%"))
                .count()
            )
            assert count == 56
            count = (
                dbsession.query(DBResource)
                .filter_by(
                    what_updated="internal-firstrun", error=None, api=None
                )
                .count()
            )
            assert count == 56
            count = (
                dbsession.query(DBResource)
                .filter_by(
                    what_updated="internal-firstrun,hash",
                    error=None,
                    api=False,
                )
                .count()
            )
            assert count == 0
            count = (
                dbsession.query(DBResource)
                .filter_by(
                    what_updated="internal-firstrun,http header,hash",
                    error=None,
                    api=False,
                )
                .count()
            )
            assert count == 0
            count = (
                dbsession.query(DBResource)
                .filter_by(what_updated="firstrun", error=None, api=None)
                .count()
            )
            assert count == 564
            count = (
                dbsession.query(DBResource)
                .filter_by(what_updated="firstrun", error=None, api=True)
                .count()
            )
            assert count == 4
            count = (
                dbsession.query(DBResource)
                .filter(DBResource.error.isnot(None))
                .filter_by(what_updated="firstrun")
                .count()
            )
            assert count == 27
            count = (
                dbsession.query(DBResource)
                .filter_by(what_updated="first hash", error=None, api=False)
                .count()
            )
            assert count == 9
            count = (
                dbsession.query(DBResource)
                .filter_by(what_updated="http header", error=None, api=None)
                .count()
            )
            assert count == 0
            count = (
                dbsession.query(DBResource)
                .filter_by(
                    what_updated="http header,hash", error=None, api=False
                )
                .count()
            )
            assert count == 0
            dbdataset = dbsession.query(DBDataset).first()
            assert (
                str(dbdataset)
                == """<Dataset(run number=0, id=a2150ad9-2b87-49f5-a6b2-c85dff366b75, reference period=09/21/2017, update frequency=1,
review date=None, last modified=2017-12-16 15:11:15.204215+00:00, metadata modified=2017-12-16 15:11:15.204215+00:00, updated by script=None,
latest of modifieds=2017-12-16 15:11:15.204215+00:00, what updated=firstrun,
Resource b21d6004-06b5-41e5-8e3e-0f28140bff64: last modified=2017-12-16 15:11:15.202742+00:00,
Dataset fresh=2, error=False)>"""
            )
            count = (
                dbsession.query(DBDataset)
                .filter_by(fresh=0, what_updated="firstrun", error=False)
                .count()
            )
            assert count == 67
            count = (
                dbsession.query(DBDataset)
                .filter_by(fresh=0, what_updated="firstrun", error=True)
                .count()
            )
            assert count == 0
            count = (
                dbsession.query(DBDataset)
                .filter_by(fresh=0, what_updated="http header", error=False)
                .count()
            )
            assert count == 0
            count = (
                dbsession.query(DBDataset)
                .filter_by(fresh=0, what_updated="http header", error=True)
                .count()
            )
            assert count == 0
            count = (
                dbsession.query(DBDataset)
                .filter_by(fresh=1, what_updated="firstrun")
                .count()
            )
            assert count == 1
            count = (
                dbsession.query(DBDataset)
                .filter_by(fresh=2, what_updated="firstrun", error=False)
                .count()
            )
            assert count == 1
            count = (
                dbsession.query(DBDataset)
                .filter_by(fresh=2, what_updated="firstrun", error=True)
                .count()
            )
            assert count == 0
            count = (
                dbsession.query(DBDataset)
                .filter_by(fresh=3, what_updated="firstrun", error=False)
                .count()
            )
            assert count == 25
            count = (
                dbsession.query(DBDataset)
                .filter_by(fresh=3, what_updated="firstrun", error=True)
                .count()
            )
            assert count == 5
            count = (
                dbsession.query(DBDataset)
                .filter_by(fresh=3, what_updated="http header")
                .count()
            )
            assert count == 0
            count = (
                dbsession.query(DBDataset)
                .filter_by(fresh=None, what_updated="firstrun", error=False)
                .count()
            )
            assert count == 4
            count = (
                dbsession.query(DBDataset)
                .filter_by(fresh=None, what_updated="firstrun", error=True)
                .count()
            )
            assert count == 0
            dbinfodataset = dbsession.query(DBInfoDataset).first()
            assert (
                str(dbinfodataset)
                == """<InfoDataset(id=a2150ad9-2b87-49f5-a6b2-c85dff366b75, name=rohingya-displacement-topline-figures, title=Rohingya Displacement Topline Figures,
private=False, organization id=hdx,
maintainer=7d7f5f8d-7e3b-483a-8de1-2b122010c1eb, location=bgd)>"""
            )
            count = dbsession.query(DBInfoDataset).count()
            assert count == 103
            dborganization = dbsession.query(DBOrganization).first()
            assert (
                str(dborganization)
                == """<Organization(id=hdx, name=hdx, title=HDX)>"""
            )
            count = dbsession.query(DBOrganization).count()
            assert count == 40



"""SQLAlchemy class representing DBResource row. Holds dynamic resource metadata for
each run.
"""
from hdx.database import Base
from sqlalchemy import Boolean, Column, ForeignKey, Integer, String

from . import CustomDateTime
from .dbinfodataset import DBInfoDataset
from .dbrun import DBRun


class DBResource(Base):
    """
    run_number = Column(
        Integer, ForeignKey(DBRun.run_number), primary_key=True
    )
    id = Column(String, primary_key=True)
    name = Column(String, nullable=False)
    dataset_id = Column(String, ForeignKey(DBInfoDataset.id), nullable=False)
    url = Column(String, nullable=False)
    last_modified = Column(CustomDateTime, nullable=False)
    metadata_modified = Column(
        CustomDateTime, default=None
    )  # this field and above are CKAN fields
    latest_of_modifieds = Column(CustomDateTime, nullable=False)
    what_updated = Column(String, nullable=False)
    http_last_modified = Column(CustomDateTime, default=None)
    md5_hash = Column(String, default=None)
    hash_last_modified = Column(CustomDateTime, default=None)
    when_checked = Column(CustomDateTime, default=None)
    api = Column(Boolean)
    error = Column(String)
    """

    run_number = Column(
        Integer, ForeignKey(DBRun.run_number), primary_key=True
    )
    id = Column(String, primary_key=True)
    name = Column(String, nullable=False)
    dataset_id = Column(String, ForeignKey(DBInfoDataset.id), nullable=False)
    url = Column(String, nullable=False)
    last_modified = Column(CustomDateTime, nullable=False)
    metadata_modified = Column(
        CustomDateTime, default=None
    )  # this field and above are CKAN fields
    latest_of_modifieds = Column(CustomDateTime, nullable=False)
    what_updated = Column(String, nullable=False)
    http_last_modified = Column(CustomDateTime, default=None)
    md5_hash = Column(String, default=None)
    hash_last_modified = Column(CustomDateTime, default=None)
    when_checked = Column(CustomDateTime, default=None)
    api = Column(Boolean)
    error = Column(String)

    def __repr__(self) -> str:
        """String representation of DBResource row

        Returns:
            str: String representation of DBResource row
        """
        output = f"<Resource(run number={self.run_number}, id={self.id}, name={self.name}, "
        output += f"dataset id={self.dataset_id},\nurl={self.url},\n"
        output += f"last modified={str(self.last_modified)}, metadata modified={str(self.metadata_modified)},\n"
        output += f"latest of modifieds={str(self.latest_of_modifieds)}, what updated={str(self.what_updated)},\n"
        output += f"http last modified={str(self.http_last_modified)},\n"
        output += f"MD5 hash={self.md5_hash}, hash last modified={str(self.hash_last_modified)}, "
        output += f"when checked={str(self.when_checked)},\n"
        output += f"api={str(self.api)}, error={str(self.error)})>"
        return output
